/*
 * UDP Session Development
 * List Utilities
 * by Joel R. Voss aka. Javantea
 * Feb 19, 2006
 */

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <string.h>
#include "list_utils.h"

// Note: You must malloc and free recv yourself unless recv is a child of 
// another recv that you will be destroying.
void recv_list_init(recv_list *recv, int32_t packet_number, BYTE *data, int data_size)
{
	recv->packet_number = packet_number;
	if(data_size <= 1024)
	{
		memcpy(recv->packet, data, data_size);
		recv->packet_size = data_size;
	}
	else
	{
		// semi-serious error here.
		strncpy(recv->packet, "Error: packet size too large.", 1024);
		recv->packet_size = -1;
	}
	// Do not dare call Memory_init on something that has already been added to!
	recv->next_recv = 0;
	//if(MEMORY_TEST) printf("Memory init: %x, val %li, %i B.\n", recv, packet_number, sizeof(recv_list));
}

void recv_list_add(recv_list *recv, recv_list *child)
{
	if(recv->next_recv == 0)
	{
		//if(MEMORY_TEST) printf("Memory add: %x, %x.\n", recv, child);
		recv->next_recv = child;
	}
	else
	{
		//if(MEMORY_TEST) printf("Memory add recurse: %x, %x.\n", recv, child);
		recv_list_add(recv->next_recv, child);
	}
}

void recv_list_add_value(recv_list *recv, int32_t packet_number, BYTE *data, int data_size)
{
	recv_list *child = (recv_list *)malloc(sizeof(recv_list));
	recv_list_init(child, packet_number, data, data_size);
	recv_list_add(recv, child);
}

void recv_list_check(const recv_list *recv, int32_t *min_out, int32_t *max_out, recv_list *missing)
{
	int32_t min = 2147483647;
	int32_t max = -1;
	recv_list_init(missing, -1, "", 0);
	//fprintf(stderr, "recv_list_check ");
	recv_list *child = recv->next_recv;
	while(child != 0)
	{
		//fprintf(stderr, "%li %i, ", child->packet_number, child->packet_size);
		if(child->packet_number < min) min = child->packet_number;
		if(child->packet_number > max) max = child->packet_number;
		child = child->next_recv;
	}
	//fprintf(stderr, "\n");
	// FIXME: This line is unstable in rare case that min = max = part:ff
	// This line Fixes the bug where min = part_0-1
	if((min & 0x000000ff) == 0x000000ff) min++;
	// Guess what the actual min and max are?
	// This is implementation-specific code.
	min &= 0xffffff00;
	max |= 0x000000ff;
	int32_t i;
	for(i = min; i < max; i++)
	{
		int found = 0;
		child = recv->next_recv;
		while(child != 0)
		{
			if(child->packet_number == i)
			{
				found = 1;
				break;
			}
			child = child->next_recv;
		}
		if(found == 0) recv_list_add_value(missing, i, "", 0);
	}
	*min_out = min;
	*max_out = max;
}

int recv_list_length(const recv_list *recv)
{
	if(recv->next_recv != 0)
	{
		return recv_list_nlength(recv->next_recv, 1);
	}
	return 0;
}

int recv_list_nlength(const recv_list *recv, int count)
{
	if(recv->next_recv != 0)
	{
		return recv_list_nlength(recv->next_recv, count + 1);
	}
	return count;
}

void recv_list_print(FILE *out_file, const recv_list *recv)
{
	fprintf(out_file, "%i ", recv->packet_number);
	if(recv->next_recv != 0)
	{
		//if(MEMORY_TEST) printf("Memory destroy: %x, %x.\n", recv, recv->next_recv);
		recv_list_print(out_file, recv->next_recv);
	}
}

void recv_list_write(const recv_list *recv, FILE *out_file)
{
	if(recv->next_recv != 0)
	{
		int32_t min, max;
		recv_list missing;
		recv_list_check(recv, &min, &max, &missing);
		int32_t i;
		for(i = min; i <= max; i++)
		{
			recv_list_nwrite(recv->next_recv, out_file, i);
		}
	}
}

void recv_list_nwrite(const recv_list *recv, FILE *out_file, int32_t i)
{
	if(recv->packet_number == i)
	{
		fwrite(recv->packet, recv->packet_size, 1, out_file);
		return;
	}
	if(recv->next_recv != 0)
	{
		recv_list_nwrite(recv->next_recv, out_file, i);
	}
	else
	{
		// This is an error condition. Sadly I have no way to mute this.
		fprintf(stderr, "Error: Unable to find this packet during write: %i\n", i);
	}
}

void recv_list_destroy(recv_list *recv)
{
	if(recv->next_recv != 0)
	{
		//if(MEMORY_TEST) printf("Memory destroy: %x, %x.\n", recv, recv->next_recv);
		recv_list_destroy(recv->next_recv);
		free(recv->next_recv);
		recv->next_recv = 0;
	}
}


// Note: You must malloc and free recv yourself unless recv is a child of 
// another recv that you will be destroying.
void recv_packet_list_init(recv_packet_list *recv, BYTE *packet, int packet_size, int32_t packet_number, struct sockaddr_in *oaddr)
{
	if(packet_size <= 1024+52)
	{
		memcpy(recv->packet, packet, packet_size);
		recv->packet_size = packet_size;
	}
	else
	{
		// semi-serious error here.
		strncpy(recv->packet, "Error: packet size too large.", 1024+52);
		recv->packet_size = -1;
	}
	recv->packet_number = packet_number;
	recv->arrival_time = time(0);
	memcpy(&(recv->oaddr), oaddr, sizeof(struct sockaddr_in));
	// Do not dare call Memory_init on something that has already been added to!
	recv->prev = 0;
	recv->next = 0;
	//if(MEMORY_TEST) printf("Memory init: %x, val %li, %i B.\n", recv, packet_number, sizeof(recv_packet_list));
}

void recv_packet_list_add(recv_packet_list *recv, recv_packet_list *child)
{
	if(recv->next == 0)
	{
		//if(MEMORY_TEST) printf("Memory add: %x, %x.\n", recv, child);
		recv->next = child;
		child->prev = recv;
	}
	else
	{
		//if(MEMORY_TEST) printf("Memory add recurse: %x, %x.\n", recv, child);
		recv_packet_list_add(recv->next, child);
	}
}

void recv_packet_list_add_value(recv_packet_list *recv, BYTE *packet, int packet_size, int32_t packet_number, struct sockaddr_in *oaddr)
{
	recv_packet_list *child = (recv_packet_list *)malloc(sizeof(recv_packet_list));
	recv_packet_list_init(child, packet, packet_size, packet_number, oaddr);
	recv_packet_list_add(recv, child);
}

recv_packet_list *recv_packet_list_search(const recv_packet_list *recv, int32_t packet_mask, int position)
{
	if(position <= 0)
	{
		//fprintf(stderr, "%08x & %08x ?= %08x : %i\n", 0xffffff00, recv->packet_number, packet_mask, (0xffffff00 & recv->packet_number) == packet_mask);
		if((0xffffff00 & recv->packet_number) == packet_mask)
		{
			return recv;
		}
	}
	if(recv != 0 && recv->next != 0)
	{
		return recv_packet_list_search(recv->next, packet_mask, position-1);
	}
	else
	{
		return 0;
	}
}

void recv_packet_list_remove(recv_packet_list *recv)
{
	if(recv->prev != 0)
	{
		// We are not the first, so connect prev and next.
		recv->prev->next = recv->next;
	}
	if(recv->next != 0)
	{
		// We are not the last, so connect prev and next.
		recv->next->prev = recv->prev;
	}
}

int recv_packet_list_length(const recv_packet_list *recv)
{
	if(recv->next != 0)
	{
		return recv_packet_list_nlength(recv->next, 0);
	}
	return 0;
}

int recv_packet_list_nlength(const recv_packet_list *recv, int count)
{
	if(recv->next != 0)
	{
		return recv_packet_list_nlength(recv->next, count + 1);
	}
	return count;
}

void recv_packet_list_print(FILE *out_file, const recv_packet_list *recv)
{
	fprintf(out_file, "%i ", recv->packet_number);
	fwrite(recv->packet, recv->packet_size, 1, stdout);
	if(recv->next != 0)
	{
		//if(MEMORY_TEST) printf("Memory destroy: %x, %x.\n", recv, recv->next_recv);
		recv_packet_list_print(out_file, recv->next);
	}
}

void recv_packet_list_destroy(recv_packet_list *recv)
{
	if(recv->next != 0)
	{
		//if(MEMORY_TEST) printf("Memory destroy: %x, %x.\n", recv, recv->next_recv);
		recv_packet_list_destroy(recv->next);
		free(recv->next);
		recv->prev = 0;
		recv->next = 0;
	}
}
